@@ -38,7 +38,7 @@
 
 #include "libdecor-plugin.h"
 #include "utils.h"
-#include "cursor-settings.h"
+#include "desktop-settings.h"
 #include "os-compatibility.h"
 
 #include <cairo/cairo.h>
@@ -322,6 +322,8 @@ struct libdecor_plugin_gtk {
 	char *cursor_theme_name;
 	int cursor_size;
 
+	uint32_t color_scheme_setting;
+
 	int double_click_time_ms;
 };
 
@@ -330,6 +332,9 @@ static const char *libdecor_gtk_proxy_tag = "libdecor-gtk";
 static bool
 own_proxy(struct wl_proxy *proxy)
 {
+	if (!proxy)
+		return false;
+
 	return (wl_proxy_get_tag(proxy) == &libdecor_gtk_proxy_tag);
 }
 
@@ -418,12 +423,17 @@ libdecor_plugin_gtk_destroy(struct libdecor_plugin *plugin)
 			free(cursor_output);
 		}
 
+		free(seat->name);
 		free(seat);
 	}
 
 	wl_list_for_each_safe(output, output_tmp,
 			      &plugin_gtk->output_list, link) {
-		wl_output_destroy(output->wl_output);
+		if (wl_output_get_version (output->wl_output) >=
+		    WL_OUTPUT_RELEASE_SINCE_VERSION)
+			wl_output_release(output->wl_output);
+		else
+			wl_output_destroy(output->wl_output);
 		free(output);
 	}
 
@@ -442,7 +452,7 @@ libdecor_plugin_gtk_destroy(struct libdecor_plugin *plugin)
 	if (plugin_gtk->wl_subcompositor)
 		wl_subcompositor_destroy(plugin_gtk->wl_subcompositor);
 
-
+	libdecor_plugin_release(&plugin_gtk->plugin);
 	free(plugin_gtk);
 }
 
@@ -587,7 +597,7 @@ create_shm_buffer(struct libdecor_plugin_gtk *plugin_gtk,
 	stride = buffer_width * 4;
 	size = stride * buffer_height;
 
-	fd = os_create_anonymous_file(size);
+	fd = libdecor_os_create_anonymous_file(size);
 	if (fd < 0) {
 		fprintf(stderr, "creating a buffer file for %d B failed: %s\n",
 			size, strerror(errno));
@@ -666,24 +676,16 @@ libdecor_plugin_gtk_frame_free(struct libdecor_plugin *plugin,
 	struct libdecor_frame_gtk *frame_gtk =
 		(struct libdecor_frame_gtk *) frame;
 
-	/* when in SSD mode, frame_gtk->header is not a proper GTK widget */
-	if (!GTK_IS_WIDGET(frame_gtk->header)) return;
-	gtk_widget_destroy(frame_gtk->header);
-	frame_gtk->header = NULL;
-	if (!GTK_IS_WIDGET(frame_gtk->window)) return;
-	gtk_widget_destroy(frame_gtk->window);
-	frame_gtk->window = NULL;
+	g_clear_pointer (&frame_gtk->header, gtk_widget_destroy);
+	g_clear_pointer (&frame_gtk->window, gtk_widget_destroy);
 
 	free_border_component(&frame_gtk->headerbar);
 	free_border_component(&frame_gtk->shadow);
 	frame_gtk->shadow_showing = false;
-	if (frame_gtk->shadow_blur != NULL) {
-		cairo_surface_destroy(frame_gtk->shadow_blur);
-		frame_gtk->shadow_blur = NULL;
-	}
 
-	free(frame_gtk->title);
-	frame_gtk->title = NULL;
+	g_clear_pointer (&frame_gtk->shadow_blur, cairo_surface_destroy);
+
+	g_clear_pointer (&frame_gtk->title, free);
 
 	frame_gtk->decoration_type = DECORATION_TYPE_NONE;
 
@@ -740,7 +742,8 @@ redraw_scale(struct libdecor_frame_gtk *frame_gtk,
 	}
 	if (scale != cmpnt->scale) {
 		cmpnt->scale = scale;
-		if ((cmpnt->type != SHADOW) || is_border_surfaces_showing(frame_gtk)) {
+		if ((frame_gtk->decoration_type != DECORATION_TYPE_NONE) &&
+		    ((cmpnt->type != SHADOW) || is_border_surfaces_showing(frame_gtk))) {
 			draw_border_component(frame_gtk, cmpnt, cmpnt->type);
 			return true;
 		}
@@ -1959,10 +1962,10 @@ component_edge(const struct border_component *cmpnt,
 	       const int pointer_y,
 	       const int margin)
 {
-	const bool top = pointer_y < margin;
-	const bool bottom = pointer_y > (cmpnt->buffer->height - margin);
-	const bool left = pointer_x < margin;
-	const bool right = pointer_x > (cmpnt->buffer->width - margin);
+	const bool top = pointer_y < margin * 2;
+	const bool bottom = pointer_y > (cmpnt->buffer->height - margin * 2);
+	const bool left = pointer_x < margin * 2;
+	const bool right = pointer_x > (cmpnt->buffer->width - margin * 2);
 
 	if (top) {
 		if (left)
@@ -2548,15 +2551,12 @@ init_wl_seat(struct libdecor_plugin_gtk *plugin_gtk,
 	struct seat *seat;
 
 	if (version < 3) {
-		char *err_msg;
-		asprintf(&err_msg,
-			 "%s version 3 required but only version %i is available\n",
-			 wl_seat_interface.name, version);
 		libdecor_notify_plugin_error(
 				plugin_gtk->context,
 				LIBDECOR_ERROR_COMPOSITOR_INCOMPATIBLE,
-				err_msg);
-		free(err_msg);
+				"%s version 3 required but only version %i is available\n",
+				wl_seat_interface.name,
+				version);
 	}
 
 	seat = zalloc(sizeof *seat);
@@ -2640,15 +2640,12 @@ init_wl_output(struct libdecor_plugin_gtk *plugin_gtk,
 	struct output *output;
 
 	if (version < 2) {
-		char *err_msg;
-		asprintf(&err_msg,
-			 "%s version 2 required but only version %i is available\n",
-			 wl_output_interface.name, version);
 		libdecor_notify_plugin_error(
 				plugin_gtk->context,
 				LIBDECOR_ERROR_COMPOSITOR_INCOMPATIBLE,
-				err_msg);
-		free(err_msg);
+				"%s version 2 required but only version %i is available\n",
+				wl_output_interface.name,
+				version);
 	}
 
 	output = zalloc(sizeof *output);
@@ -2657,7 +2654,8 @@ init_wl_output(struct libdecor_plugin_gtk *plugin_gtk,
 	output->id = id;
 	output->wl_output =
 		wl_registry_bind(plugin_gtk->wl_registry,
-				 id, &wl_output_interface, 2);
+				 id, &wl_output_interface,
+				 MIN (version, 3));
 	wl_proxy_set_tag((struct wl_proxy *) output->wl_output,
 			 &libdecor_gtk_proxy_tag);
 	wl_output_add_listener(output->wl_output, &output_listener, output);
@@ -2777,6 +2775,12 @@ libdecor_plugin_new(struct libdecor *context)
 	struct libdecor_plugin_gtk *plugin_gtk;
 	struct wl_display *wl_display;
 
+#ifdef HAVE_GETTID
+	/* Only support running on the main thread. */
+	if (getpid () != gettid ())
+		return NULL;
+#endif
+
 	plugin_gtk = zalloc(sizeof *plugin_gtk);
 	libdecor_plugin_init(&plugin_gtk->plugin,
 			     context,
@@ -2794,6 +2798,8 @@ libdecor_plugin_new(struct libdecor *context)
 		plugin_gtk->cursor_size = 24;
 	}
 
+	plugin_gtk->color_scheme_setting = libdecor_get_color_scheme();
+
 	wl_display = libdecor_get_wl_display(context);
 	plugin_gtk->wl_registry = wl_display_get_registry(wl_display);
 	wl_registry_add_listener(plugin_gtk->wl_registry,
@@ -2814,14 +2820,19 @@ libdecor_plugin_new(struct libdecor *context)
 
 	/* setup GTK context */
 	gdk_set_allowed_backends("wayland");
+	gtk_disable_setlocale();
+
 	if (!gtk_init_check(NULL, NULL)) {
-		libdecor_notify_plugin_error(
-				plugin_gtk->context,
-				LIBDECOR_ERROR_COMPOSITOR_INCOMPATIBLE,
-				"GTK cannot connect to Wayland compositor");
+		fprintf(stderr, "libdecor-gtk-WARNING: Failed to initialize GTK\n");
+		libdecor_plugin_gtk_destroy(&plugin_gtk->plugin);
 		return NULL;
 	}
 
+	g_object_set(gtk_settings_get_default(),
+		     "gtk-application-prefer-dark-theme",
+		     plugin_gtk->color_scheme_setting == LIBDECOR_COLOR_SCHEME_PREFER_DARK,
+		     NULL);
+
 	return &plugin_gtk->plugin;
 }
 
@@ -2838,6 +2849,7 @@ libdecor_plugin_description = {
 	.constructor = libdecor_plugin_new,
 	.conflicting_symbols = {
 		"png_free",
+		"gdk_get_use_xshm",
 		NULL,
 	},
 };
